@@ -1,7 +1,7 @@
 /*
  *  SPDX-License-Identifier: GPL-2.0-or-later
  *
- *  Copyright (C) 2020-2022  The DOSBox Staging Team
+ *  Copyright (C) 2020-2023  The DOSBox Staging Team
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -21,6 +21,7 @@
 #ifndef DOSBOX_PROGRAM_LS_H
 #define DOSBOX_PROGRAM_LS_H
 
+#include "dos_inc.h"
 #include "programs.h"
 
 #include <string>
@@ -29,12 +30,23 @@ class LS final : public Program {
 public:
 	LS()
 	{
+		AddMessages();
 		help_detail = {HELP_Filter::All,
 		               HELP_Category::File,
 		               HELP_CmdType::Program,
 		               "LS"};
 	}
-	void Run();
+	void Run() override;
+
+private:
+	// Map a vector of dir contents to a vector of word widths
+	static std::vector<uint8_t> GetFileNameLengths(
+	        const std::vector<DOS_DTA::Result>& dir_contents,
+	        const uint8_t padding = 0);
+	static std::vector<uint8_t> GetColumnWidths(const std::vector<uint8_t>& name_widths,
+	        const uint8_t min_column_width);
+
+	static void AddMessages();
 };
 
 #endif
